#ifndef COCO_RUNTIME_H_
#define COCO_RUNTIME_H_

// MIT License
//
// Copyright (c) 2019 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include <algorithm>
#include <array>
#include <cassert>
#include <chrono>
#include <condition_variable>
#include <deque>
#include <map>
#include <new>
#include <string>
#include <type_traits>
#include <utility>
#include <vector>

#if !defined(__clang__) && defined(__GNUC__) && (__GNUC__ < 4 || (__GNUC__ == 4 && __GNUC_MINOR__ < 5))
#define COCO_GCC44_WORKAROUNDS
#endif

#if !defined(__clang__) && defined(__GNUC__) && (__GNUC__ < 4 || (__GNUC__ == 4 && __GNUC_MINOR__ < 9))
#define COCO_GCC48_WORKAROUNDS
#include <boost/integer/static_min_max.hpp>
#include <boost/type_traits/aligned_storage.hpp>
#endif

#ifndef COCO_GCC44_WORKAROUNDS
#include <atomic>
#endif

#if (__cplusplus >= 201703L) || (defined(_MSVC_LANG) && (_MSVC_LANG >= 201703L) && (_MSC_VER >= 1913))
#include <optional>
#include <variant>
#else
#define COCO_VARIANT_WORKAROUND
#endif

namespace coco {
inline namespace v1 {
#ifdef COCO_GCC48_WORKAROUNDS
namespace detail {

struct EmptyType {};

template <size_t... Args>
struct max;

template <>
struct max<> {
  static const size_t value = 0;
};

template <size_t T, size_t... Ts>
struct max<T, Ts...> {
  static const size_t value = boost::static_unsigned_max<T, max<Ts...>::value>::value;
};

}  // namespace detail

template <size_t lb, class... Args>
struct aligned_union {
  typedef typename boost::aligned_storage<boost::static_unsigned_max<lb, detail::max<sizeof(Args)...>::value>::value,
                                          detail::max<boost::alignment_of<Args>::value...>::value>::type type;
};
#else
template <size_t lb, typename... Args>
using aligned_union = std::aligned_union<lb, Args...>;
#endif

#ifdef COCO_GCC44_WORKAROUNDS
typedef std::chrono::monotonic_clock CocoClock;
#else
typedef std::chrono::steady_clock CocoClock;
#endif

#ifdef COCO_GCC44_WORKAROUNDS
#define COCO_NOEXCEPT
#define COCO_OVERRIDE
#else
#define COCO_NOEXCEPT noexcept
#define COCO_OVERRIDE override
#endif

/// The version of the runtime this file implements.
const unsigned int kRuntimeVersion = 5;

class BaseComponent;
class BaseSingleThreadedComponent;
class StateMachineLogger;

#ifdef COCO_RUNTIME_ASSERT_ENABLE
#define COCO_RUNTIME_ASSERT(cond) assert(cond)
#else
#define COCO_RUNTIME_ASSERT(cond) ((void)0)
#endif

#ifdef NDEBUG
#define COCO_ALWAYS_ASSERT(cond) \
  if (!(cond)) {                 \
    abort();                     \
  }
#else
#define COCO_ALWAYS_ASSERT(cond) assert(cond)
#endif

/// A small stl-compatible (in so far as the methods are available) set implementation that uses a vector.
template <typename T>
class vector_set {
 public:
  typedef typename std::vector<T>::iterator iterator;
  typedef typename std::vector<T>::const_iterator const_iterator;
  typedef typename std::vector<T>::size_type size_type;

  vector_set() = default;

  const_iterator begin() const { return items_.begin(); }
  const_iterator end() const { return items_.end(); }

  std::pair<iterator, bool> insert(T&& item) {
    iterator it = std::find(items_.begin(), items_.end(), item);
    if (it != items_.end()) {
      return std::make_pair(it, false);
    }
    items_.emplace_back(std::move(item));
    return std::make_pair(items_.end() - 1, true);
  }

  size_type erase(const T& item) {
    iterator it = std::find(items_.begin(), items_.end(), item);
    if (it == items_.end()) {
      return 0;
    }
    std::swap(*it, items_.back());
    items_.pop_back();
    return 1;
  }

 private:
  std::vector<T> items_;
};

/// The base class of all Coco ports.
class BasePort {
 public:
  BasePort() {}

#ifndef COCO_GCC44_WORKAROUNDS
  BasePort(const BasePort&) = delete;
  BasePort& operator=(const BasePort&) = delete;
#endif

  virtual ~BasePort() = default;

  /// Starts the component's state machine.
  ///
  /// @sa BaseComponent::coco_start
  virtual void coco_start() {}
};

class SingleThreadedSubscriber {
 public:
  virtual ~SingleThreadedSubscriber() = default;

  /// For components using the single-threaded runtime, this will cause them to process all items in their queues.
  ///
  /// The component will process its queue through to completion. The component will then recursively call
  /// coco_drain_queue() on its provided port. This means that the queues will be recursively drained upwards through a
  /// single-threaded component stack.
  virtual void coco_drain_queue() = 0;
};

class BaseProvidedPort : public BasePort {
 public:
  virtual void single_threaded_subscribe(SingleThreadedSubscriber& port) {
    // If this abort is executed, this means that the component providing this port was expected to be single-threaded,
    // but is instead multi-threaded.
    (void)port;
    abort();
  }
  virtual void single_threaded_unsubscribe(SingleThreadedSubscriber& port) {
    // If this abort is executed, this means that the component providing this port was expected to be single-threaded,
    // but is instead multi-threaded.
    (void)port;
    abort();
  }
};

/// Represents something that requires connecting to a provided port.
///
/// The most common use case of a ConnectionPoint is a required port on an implementation component which requires
/// connecting before the system is started. ConnectionPoints are also used on encapsulating components, in which case
/// calling connect will connect all internal users.
template <class ProvidedPort>
class ConnectionPoint {
 public:
  /// Connects all users of this connection point to the specified port.
  ///
  /// @param port The port to connect to.
  /// @param should_subscribe Indicates if all internal users should be subscribed to signals from port.
  virtual void connect(ProvidedPort& port, bool should_subscribe = true) = 0;
};

class ComponentObserver {
 public:
  virtual ~ComponentObserver() = default;

  /// Called when a state machine is about to be started.
  virtual void machine_starting(const BaseComponent& component) { (void)component; }

  /// Called when a state machine has succesfully shutdown.
  ///
  /// In all runtimes, this is called after StateMachineLogger::state_exit_finished() has been called for the
  /// `Terminated` state. Otherwise:
  /// * In the single-threaded runtime, this is called just before the event that caused the transition to `Terminated`
  ///   is executed.
  /// * In the multi-threaded runtime, this is called as the final action of the background thread if there is one. If
  ///   there is no background thread (because the component does not handle signals, or have any execution states),
  ///   then this will be called from the handler of the function call that caused the transition to Terminated.
  virtual void machine_terminated(const BaseComponent& component) { (void)component; }
};

/// The base class of all Coco components.
class BaseComponent {
 public:
  BaseComponent() : coco_logger_(NULL), started_(false) {}
  virtual ~BaseComponent() = default;

  BaseComponent(const BaseComponent&) = delete;
  BaseComponent& operator=(const BaseComponent&) = delete;

  /// A name for this component.
  ///
  /// By default this will be set to the name of the type from the Coco file or, if this component is owned by a Coco
  /// encapsulating component, to the name of the owner component.
  const std::string& name() const { return name_; }
  void set_name(const std::string& name) { name_ = name; }
  void set_name(const std::string& base_name, const std::string& field_name) {
    name_.clear();
    name_.reserve(base_name.size() + 1 + field_name.size());
    name_.append(base_name);
    name_.push_back('.');
    name_.append(field_name);
  }

  /// Returns the current logger that is used by the machine.
  StateMachineLogger* logger() const { return coco_logger_; }

  /// Sets the logger that should be used by the machine.
  ///
  /// This should not be used once start() has been called.
  void set_logger(StateMachineLogger* logger) {
    COCO_RUNTIME_ASSERT(!started_ && "loggers can only be set before starting a state machine");
    coco_logger_ = logger;
  }

  /// Sets the logger that should be used by this state machine, and all state machines owned by this component.
  ///
  /// In order to enable the logging, generator.cpp.alwaysEnableLogging must be true (by default this setting is true),
  /// or the generated code must be compiled with COCO_LOGGING_ENABLE defined..
  ///
  /// This should not be used once start() has been called. The component will hold a reference to the passed logger;
  /// the caller is responsible for ensuring that the logger is kept alive until the state machine has terminated.
  virtual void set_logger_recursively(StateMachineLogger* logger) { set_logger(logger); }

  /// Starts the component's state machine.
  ///
  /// After this call has returned, it is safe to call all of the methods on this component's provided port. This may be
  /// called more than once in which case subsequent calls will be ignored.
  virtual void coco_start() = 0;

  /// Adds an observer who will be notified about key component lifecycle events.
  virtual void add_observer(ComponentObserver& observer) { observers_.push_back(&observer); }

 protected:
  std::string name_;
  StateMachineLogger* coco_logger_;
  bool started_;

  /// Informs all registered ComponentObserver that this machine is starting.
  ///
  /// Do not call this directly.
  void send_machine_starting() {
    for (auto it = observers_.begin(), it_end = observers_.end(); it != it_end; ++it) {
      (*it)->machine_starting(*this);
    }
  }

  /// Informs all registered ComponentObserver that this machine has terminated.
  ///
  /// Do not call this directly.
  void send_machine_terminated() {
    for (auto it = observers_.begin(), it_end = observers_.end(); it != it_end; ++it) {
      (*it)->machine_terminated(*this);
    }
  }

 private:
  std::vector<ComponentObserver*> observers_;
};

/// An abstract interface to log events from a Coco state machine.
class StateMachineLogger : public ComponentObserver {
 public:
  virtual ~StateMachineLogger() = default;

  /// Called after a state machine has been started and its entry action fully executed.
  virtual void machine_started(const BaseComponent& component, const char* current_state) {
    (void)component;
    (void)current_state;
  }

  /// Call whenever a component's state machine enters a state.
  ///
  /// This called just *before* the body of the entry action is executed.
  virtual void state_entry_started(const BaseComponent& component, const char* state) {
    (void)component;
    (void)state;
  }

  /// Call whenever a component's state machine exits a state.
  ///
  /// This called just *after* the body of the exit action is executed.
  virtual void state_exit_finished(const BaseComponent& component, const char* state) {
    (void)component;
    (void)state;
  }

  /// Called when an event is about to be processed by the state machine, but before a handler has been chosen.
  ///
  /// @param component The component whose state machine is processing the event.
  /// @param current_state The string name of the current state of the state machine.
  /// @param port The string name corresponding to the port from which the event was received.
  /// @param event The string name of the event being received.
  virtual void event_started(const BaseComponent& component, const char* current_state, const char* port,
                             const char* event) {
    (void)component;
    (void)current_state;
    (void)port;
    (void)event;
  }

  /// Called after the event has been processed to completion.
  ///
  /// This includes after all relevant entry/exit actions have been fired. Further, in the single-threaded model it
  /// will only be fired after the queue has finished being drained.
  virtual void event_finished(const BaseComponent& component, const char* current_state, const char* port,
                              const char* event) {
    (void)component;
    (void)current_state;
    (void)port;
    (void)event;
  }

  /// Called whenever a call to Coco's log() function is executed.
  ///
  /// @param component The component whose state machine is processing the event.
  /// @param current_state The string name of the current state of the state machine.
  /// @param message The formatted message.
  virtual void message(const BaseComponent& component, const char* current_state, const char* message) {
    (void)component;
    (void)current_state;
    (void)message;
  }

  /// Called whenever a call to Coco's abort() function is executed in a state machine.
  ///
  /// Subclasses do not need to call abort: the calling component will immediately call abort afterwards instead.
  ///
  /// @param component The component whose state machine is aborting.
  /// @param current_state The string name of the current state of the state machine.
  virtual void abort(const BaseComponent& component, const char* current_state) {
    (void)component;
    (void)current_state;
  }
};

/// \defgroup AlwaysLogging Selective logging macros
///
/// @{

#define COCO_LOG(method, component, ...)                        \
  if ((component).coco_logger_ != NULL) {                       \
    (component).coco_logger_->method(component, ##__VA_ARGS__); \
  }

#define COCO_LOG_STATE_MACHINE_STARTING(component) COCO_LOG(machine_starting, component)
#define COCO_LOG_STATE_MACHINE_STARTED(component, state) COCO_LOG(machine_started, component, state)
#define COCO_LOG_STATE_MACHINE_TERMINATED(component) COCO_LOG(machine_terminated, component)
#define COCO_LOG_STATE_ENTRY_STARTED(component, state) COCO_LOG(state_entry_started, component, state)
#define COCO_LOG_STATE_EXIT_FINISHED(component, state) COCO_LOG(state_exit_finished, component, state)
#define COCO_LOG_EVENT_STARTED(component, state, port, event) COCO_LOG(event_started, component, state, port, event)
#define COCO_LOG_EVENT_FINISHED(component, state, port, event) COCO_LOG(event_finished, component, state, port, event)
#define COCO_LOG_USER(component, state, msg) COCO_LOG(message, component, state, msg)
#define COCO_LOG_ABORT(component, state) \
  COCO_LOG(abort, component, state)      \
  abort();

/// @}

/// \defgroup SelectiveLogging Selective logging macros
///
/// These macros allow logging to be disabled or enabled via a define, COCO_LOGGING_ENABLE.
///
/// @{

#if defined(COCO_LOGGING_ENABLE)
#define COCO_MAYBE_LOG(method, component, ...) COCO_LOG(method, component, ##__VA_ARGS__);
#else
#define COCO_MAYBE_LOG(...)
#endif

#define COCO_MAYBE_LOG_STATE_MACHINE_STARTING(component) COCO_MAYBE_LOG(machine_starting, component)
#define COCO_MAYBE_LOG_STATE_MACHINE_STARTED(component, state) COCO_MAYBE_LOG(machine_started, component, state)
#define COCO_MAYBE_LOG_STATE_MACHINE_TERMINATED(component) COCO_MAYBE_LOG(machine_terminated, component)
#define COCO_MAYBE_LOG_STATE_ENTRY_STARTED(component, state) COCO_MAYBE_LOG(state_entry_started, component, state)
#define COCO_MAYBE_LOG_STATE_EXIT_FINISHED(component, state) COCO_MAYBE_LOG(state_exit_finished, component, state)
#define COCO_MAYBE_LOG_EVENT_STARTED(component, state, port, event) \
  COCO_MAYBE_LOG(event_started, component, state, port, event)
#define COCO_MAYBE_LOG_EVENT_FINISHED(component, state, port, event) \
  COCO_MAYBE_LOG(event_finished, component, state, port, event)
#define COCO_MAYBE_LOG_USER(component, state, msg) COCO_MAYBE_LOG(message, component, state, msg)
#define COCO_MAYBE_LOG_ABORT(component, state) \
  COCO_MAYBE_LOG(abort, component, state)      \
  abort();

/// @}

/// An abstract class that can be implemented to provide an implementation of timers in the single-threaded runtime.
class SingleThreadedTimerExecutor {
 public:
  SingleThreadedTimerExecutor() = default;
  virtual ~SingleThreadedTimerExecutor() = default;

  SingleThreadedTimerExecutor(const SingleThreadedTimerExecutor&) = delete;
  SingleThreadedTimerExecutor& operator=(const SingleThreadedTimerExecutor&) = delete;

  /// Enables a timer.
  ///
  /// When the timer expires, the timer implementation should call execute_timer on the component with the given
  /// timer_id. Note that there is no locking provided by the single-threaded implementation: it is the responsibility
  /// of the implementor of this class to ensure that it prevents any function calls or signals from being executed
  /// whilst the timer itself is executing.
  virtual void enable_timer(BaseSingleThreadedComponent& component, int timer_id, CocoClock::duration duration) = 0;

  /// Cancels a timer.
  ///
  /// After this call, the timer implementation must not call execute_timer on the given timer until it has been
  /// re-enabled. Note that this includes cancelling any pending execute_timer calls that have been scheduled but have
  /// not yet been dispatched.
  virtual void cancel_timer(BaseSingleThreadedComponent& component, int timer) = 0;
};

/// The base class of all components that use the single-threaded runtime.
class BaseSingleThreadedComponent : public BaseComponent, public SingleThreadedSubscriber {
 public:
  BaseSingleThreadedComponent() : timer_executor_(NULL) {}

  /// Sets the timer executor to be used to execute any timers that this single-threaded component has.
  ///
  /// This does not normally need to be called: it is automatically called by the runtime when reuqired.
  ///
  /// For components that use timers this must be called prior to BaseComponent::coco_start().
  virtual void set_timer_executor(SingleThreadedTimerExecutor& executor) { timer_executor_ = &executor; }

  /// Executes the given timer.
  ///
  /// This does not normally need to be called: it is automatically called by the runtime when reuqired.
  ///
  /// This method should **only** be called by this component's SingleThreadedTimerExecutor.
  virtual void execute_timer(int timer_id) {
    (void)timer_id;
    abort();
  }

  void coco_drain_queue() COCO_OVERRIDE { send_drain_queue(); }

 protected:
  void cancel_timer(int timer_id) {
    COCO_RUNTIME_ASSERT(timer_executor_);
    timer_executor_->cancel_timer(*this, timer_id);
  }

  void enable_timer(int timer_id, CocoClock::duration duration) {
    COCO_RUNTIME_ASSERT(timer_executor_);
    timer_executor_->enable_timer(*this, timer_id, duration);
  }

  /// Drains the queue of all subscribers
  void send_drain_queue() {
    for (auto it = single_threaded_subscribers_.begin(), it_end = single_threaded_subscribers_.end(); it != it_end;
         ++it) {
      (*it)->coco_drain_queue();
    }
  }

  /// Subscribe to drain queue calls from this component.
  void add_single_threaded_subscriber(SingleThreadedSubscriber& subscriber) {
    single_threaded_subscribers_.insert(&subscriber);
  }

  /// Unsubscribe drain queue calls from this component.
  void remove_single_threaded_subscriber(SingleThreadedSubscriber& subscriber) {
    single_threaded_subscribers_.erase(&subscriber);
  }

 private:
  SingleThreadedTimerExecutor* timer_executor_;
  vector_set<SingleThreadedSubscriber*> single_threaded_subscribers_;
};

/// An encapsulating component that contains single-threaded components.
class BaseSingleThreadedEncapsulatingComponent : public BaseComponent, protected ComponentObserver {
 public:
  BaseSingleThreadedEncapsulatingComponent() : running_machines_(0) {}
  ~BaseSingleThreadedEncapsulatingComponent() {
    // assert(running_machines_ == 0);
  }

 protected:
  bool have_all_machines_terminated() const { return running_machines_ == 0; }

  virtual void all_machines_terminated() {
    COCO_LOG_STATE_MACHINE_TERMINATED(*this);
    send_machine_terminated();
  }

 private:
  unsigned int running_machines_;

  void machine_starting(const BaseComponent&) COCO_OVERRIDE {
    if (++running_machines_ == 1) {
      send_machine_starting();
    }
  }

  void machine_terminated(const BaseComponent&) COCO_OVERRIDE {
    COCO_RUNTIME_ASSERT(running_machines_ > 0);
    if (--running_machines_ == 0) {
      all_machines_terminated();
    }
  }
};

/// An encapsulating component that contains multi-threaded components.
class BaseMultiThreadedEncapsulatingComponent : public BaseComponent, protected ComponentObserver {
 public:
  BaseMultiThreadedEncapsulatingComponent() : running_machines_(0) {}
  ~BaseMultiThreadedEncapsulatingComponent() {}

 protected:
  std::mutex mutex_;

  /// Have all machines already stopped?
  ///
  /// Requires: mutex_ locked
  bool have_all_machines_terminated() const { return running_machines_ == 0; }

  void wait_for_component_termination() {
    std::unique_lock<std::mutex> lock(mutex_);
    while (!have_all_machines_terminated()) {
      all_machines_terminated_.wait(lock);
    }
  }

 private:
  std::condition_variable all_machines_terminated_;
  unsigned int running_machines_;

  void machine_starting(const BaseComponent&) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex_);
    if (++running_machines_ == 1) {
      send_machine_starting();
    }
  }

  void machine_terminated(const BaseComponent&) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex_);
    COCO_RUNTIME_ASSERT(running_machines_ > 0);
    if (--running_machines_ == 0) {
      COCO_LOG_STATE_MACHINE_TERMINATED(*this);
      send_machine_terminated();
      all_machines_terminated_.notify_all();
    }
  }
};

/// A fixed-size queue implemented using a circular buffer.
///
/// @tparam T the type of items in the queue. Must be either movable or copyable.
/// @tparam capactiy the maximum allowed size of the queue
template <class T, std::size_t capacity>
class FixedSizeCircularBufferQueue {
 public:
  typedef std::size_t size_type;

  /// Creates a new empty circular buffer.
  FixedSizeCircularBufferQueue() { start_ = size_ = 0; }

  FixedSizeCircularBufferQueue(const FixedSizeCircularBufferQueue&) = delete;
  FixedSizeCircularBufferQueue& operator=(const FixedSizeCircularBufferQueue&) = delete;

  ~FixedSizeCircularBufferQueue() {
    // Destroy all items left in the queue
    for (size_type i = 0; i < size_; ++i) {
      item((start_ + i) % capacity)->~T();
    }
  }

  bool empty() const { return size_ == 0; }

  T pop() {
    COCO_ALWAYS_ASSERT(size_ > 0 && "queue empty when pop() called");
    T result{std::move(reinterpret_cast<T&&>(data_[start_]))};
    item(start_)->~T();
    if (++start_ == capacity) {
      start_ = 0;
    }
    --size_;
    return result;
  }

  void push(T&& new_item) {
    COCO_ALWAYS_ASSERT(size_ < capacity && "fixed-size queue full");
    ::new (item((start_ + size_) % capacity)) T(std::move(new_item));
    ++size_;
  }

  /// Does the queue contain an item that satisfies the given predicate?
  template <class UnaryPredicate>
  void push_unique(T&& new_item, UnaryPredicate predicate) {
    size_type i = start_;
    for (size_type considered = 0; considered < size_; ++considered) {
      if (predicate(*item(i))) {
        // Replace
        *item(i) = std::move(new_item);
        return;
      }
      if (++i == capacity) {
        i = 0;
      }
    }
    push(std::move(new_item));
  }

  /// Removes all items that satisfy the predicate.
  template <class UnaryPredicate>
  void purge(UnaryPredicate predicate) {
    size_type destination = start_;
    size_type source = start_;
    size_type removed = 0;
    for (size_type i = 0; i < size_; ++i) {
      if (predicate(*item(source))) {
        ++removed;
      } else {
        if (source != destination) {
          // Shift items down
          *item(destination) = std::move(*item(source));
        }
        if (++destination == capacity) {
          destination = 0;
        }
      }
      if (++source == capacity) {
        source = 0;
      }
    }
    // Destroy discarded objects at the end
    for (size_type i = 0; i < removed; ++i) {
      item(destination++)->~T();
      if (destination == capacity) {
        destination = 0;
      }
    }
    size_ -= removed;
  }

  /// Does the queue contain an item that satisfies the given predicate?
  template <class UnaryPredicate>
  bool contains(UnaryPredicate predicate) const {
    size_type i = start_;
    for (size_type considered = 0; considered < size_; ++considered) {
      if (predicate(*item(i))) {
        return true;
      }
      if (++i == capacity) {
        i = 0;
      }
    }
    return false;
  }

  /// Clears all items in the queue.
  void clear() {
    for (size_type i = 0; i < size_; ++i) {
      item((start_ + i) % capacity)->~T();
    }
    start_ = size_ = 0;
  }

 private:
#ifdef COCO_GCC44_WORKAROUNDS
  typename boost::aligned_storage<sizeof(T), boost::alignment_of<T>::value>::type data_[capacity];
#else
  alignas(T) std::array<unsigned char, sizeof(T)> data_[capacity];
#endif
  size_type start_;
  size_type size_;

  T* item(size_type i) { return reinterpret_cast<T*>(&data_[i]); }
  const T* item(size_type i) const { return reinterpret_cast<const T*>(&data_[i]); }
};

/// A variable-size queue implemented using a circular buffer.
///
/// @tparam T the type of items in the queue. Must be either movable or copyable.
template <class T>
class VariableSizeQueue {
 public:
  VariableSizeQueue() {}

  VariableSizeQueue(const VariableSizeQueue&) = delete;
  VariableSizeQueue& operator=(const VariableSizeQueue&) = delete;

  bool empty() const { return queue_.empty(); }

  T pop() {
    T result{std::move(queue_.front())};
    queue_.pop_front();
    return result;
  }

  void push(T&& new_item) { queue_.push_back(std::move(new_item)); }

  /// Does the queue contain an item that satisfies the given predicate?
  template <class UnaryPredicate>
  void push_unique(T&& new_item, UnaryPredicate predicate) {
    for (typename std::deque<T>::iterator it = queue_.begin(), it_end = queue_.end(); it != it_end; ++it) {
      if (predicate(*it)) {
        *it = std::move(new_item);
        return;
      }
    }
    push(std::move(new_item));
  }

  /// Removes all items that satisfy the predicate.
  template <class UnaryPredicate>
  void purge(UnaryPredicate predicate) {
    typename std::deque<T>::size_type destination = 0, source = 0;
    for (typename std::deque<T>::size_type i = 0; i < queue_.size(); ++i) {
      if (!predicate(queue_[source])) {
        if (source != destination) {
          // Shift items down
          queue_[destination] = std::move(queue_[source]);
        }
        ++destination;
      }
      ++source;
    }
    queue_.erase(queue_.begin() + static_cast<typename std::deque<T>::difference_type>(destination), queue_.end());
  }

  /// Does the queue contain an item that satisfies the given predicate?
  template <class UnaryPredicate>
  bool contains(UnaryPredicate predicate) const {
    return std::any_of(queue_.begin(), queue_.end(), predicate);
  }

  /// Clears all items in the queue.
  void clear() { queue_.clear(); }

 private:
  std::deque<T> queue_;
};

/// Represents the Coco Range type, which is a half-open interval `[lower, upper)`.
template <typename Integer = int>
struct Range : std::pair<Integer, Integer> {
  class iterator {
   public:
    typedef Integer value_type;
    typedef std::make_signed<Integer> difference_type;
    typedef std::random_access_iterator_tag iterator_category;

    explicit iterator(Integer value) : value_(value) {}

    bool operator<(const iterator& other) const { return value_ < other.value_; }
    bool operator<=(const iterator& other) const { return value_ <= other.value_; }
    bool operator>(const iterator& other) const { return value_ > other.value_; }
    bool operator>=(const iterator& other) const { return value_ >= other.value_; }
    bool operator==(const iterator& other) const { return value_ == other.value_; }
    bool operator!=(const iterator& other) const { return value_ != other.value_; }

    value_type operator*() const { return value_; }

    iterator operator+(Integer increment) const { return iterator{value_ + increment}; }
    iterator operator-(Integer increment) const { return iterator{value_ - increment}; }
    iterator& operator+=(Integer increment) {
      value_ += increment;
      return *this;
    }
    iterator& operator-=(Integer increment) {
      value_ -= increment;
      return *this;
    }
    void operator++() { ++value_; }

    difference_type operator-(const iterator& other) const { return value_ - other.value_; }

   private:
    Integer value_;
  };

  typedef iterator const_iterator;

  Range(Integer first, Integer second) : std::pair<Integer, Integer>(first, second) {}
  Range(const Range& range) : std::pair<Integer, Integer>(range) {}

  iterator begin() const { return iterator{std::pair<Integer, Integer>::first}; }
  iterator end() const { return iterator{std::pair<Integer, Integer>::second}; }
};

namespace detail {

/// A queue of timers, each with their own independent deadlines.
///
/// This can be used to implement a set of timers using a single worker thread.
///
/// @tparam TimerId a type that is used to represent the type of a timer. It must be statically castable to an int.
/// @tparam Clock the clock to use for mesurement of time periods
/// @tparam QueueDetail a class used to implement the queue; see FixedSizeTimerQueueConfig for an example.
template <class TimerId, class Clock, class QueueDetail>
class TimerQueue {
 public:
  typedef typename Clock::duration Duration;

  TimerQueue() : heap_size_(0), enabled_(0) {}

  TimerQueue(const TimerQueue&) = delete;
  TimerQueue& operator=(const TimerQueue&) = delete;

  /// Returns true if any timers are enabled at the moment.
  bool timer_enabled() const { return enabled_ > 0; }

  /// Assuming a timer is enabled, computes the next time point at which one of these timers will fire.
  std::chrono::time_point<Clock> next_deadline() {
    COCO_RUNTIME_ASSERT(timer_enabled());
    while (!timer_heap_[0]->enabled) {
      // Just remove from the heap
      timer_heap_[0]->in_heap = false;
      std::pop_heap(timer_heap_.begin(), timer_heap_.begin() + heap_size_, heap_compare_timers);
      --heap_size_;
      COCO_RUNTIME_ASSERT(heap_size_ > 0);
    }
    return timer_heap_[0]->next_deadline;
  }

  bool timer_fired() { return timer_enabled() && next_deadline() <= Clock::now(); }

  /// Causes the current thread to block until the next deadline is reached, if relevant.
  template <class ConditionVariable, class Lock>
  void wait_until_next_deadline(ConditionVariable& condition_variable, Lock& lock, bool handler_running) {
    if (!timer_enabled()) {
      condition_variable.wait(lock);
    } else {
      std::chrono::time_point<Clock> target = next_deadline();
      if (target <= Clock::now()) {
        if (handler_running) {
          condition_variable.wait(lock);
        }
      } else {
        condition_variable.wait_until(lock, target);
      }
    }
  }

  /// Assuming a timer is enabled, returns the first timer whose deadline as expired.
  ///
  /// @pre next_deadline() <= now().
  TimerId pop() {
    COCO_RUNTIME_ASSERT(timer_fired());

    // Remove from the heap
    std::pop_heap(timer_heap_.begin(), timer_heap_.begin() + heap_size_, heap_compare_timers);
    --heap_size_;

    Timer* popped_timer = timer_heap_[heap_size_];
    COCO_RUNTIME_ASSERT(popped_timer->in_heap && popped_timer->enabled);
    popped_timer->in_heap = false;
    popped_timer->enabled = false;
    --enabled_;

    return popped_timer->id;
  }

  /// Enables a timer, setting its kind and duration.
  ///
  /// Enabling an already-enabled timer has no effect.
  void enable_timer(TimerId timer_id, Duration duration) {
    Timer* timer = get_timer(timer_id);
    if (timer->enabled) {
      return;
    }
    timer->id = timer_id;
    timer->enabled = true;
    timer->enable_order = heap_size_;
    timer->duration = duration;
    timer->next_deadline = Clock::now() + duration;

    if (timer->in_heap) {
      // Already in the heap; need to modify its position
      // TODO: improve performance
      std::make_heap(timer_heap_.begin(), timer_heap_.begin() + heap_size_, heap_compare_timers);
    } else {
      // Insert into the heap
      timer->in_heap = true;
      timer_heap_[heap_size_] = timer;
      ++heap_size_;
      std::push_heap(timer_heap_.begin(), timer_heap_.begin() + heap_size_, heap_compare_timers);
    }

    ++enabled_;
  }

  /// Cancels the given timer.
  ///
  /// Cancelling an already cancelled timer has no effect.
  void cancel_timer(TimerId timer_id) {
    Timer* timer = get_timer(timer_id);
    if (!timer->enabled) {
      return;
    }
    // We don't remove from the queue here - next_deadline will simply walk over disabled timers
    timer->enabled = false;
    --enabled_;
  }

 private:
  struct Timer {
    Timer() : enabled(false), in_heap(false) {}
    TimerId id;
    unsigned int enable_order;
    bool enabled;
    bool in_heap;
    Duration duration;
    std::chrono::time_point<Clock> next_deadline;
  };

  typename QueueDetail::template Container<Timer>::type timers_;
  typename QueueDetail::template Container<Timer*>::type timer_heap_;
  unsigned int heap_size_;
  unsigned int enabled_;

  Timer* get_timer(TimerId timer) {
    return QueueDetail::template get_timer<Timer>(timers_, timer_heap_, static_cast<unsigned int>(timer));
  }

  /// Compares timer first by deadline, then by their id to ensure consistent determinsitic ordering.
  static bool heap_compare_timers(const Timer* left, const Timer* right) {
    return left->next_deadline > right->next_deadline ||
           (left->next_deadline == right->next_deadline && left->enable_order > right->enable_order);
  }
};

/// A QueueConfig for TimerQueue that uses a fixed-capacity array.
template <std::size_t capacity>
struct FixedSizeTimerQueueConfig {
  template <class T>
  struct Container {
    typedef std::array<T, capacity> type;
  };

  template <class Timer>
  static Timer* get_timer(typename Container<Timer>::type& timers, typename Container<Timer*>::type&,
                          unsigned int timer) {
    return &timers[timer];
  }
};

/// A QueueConfig for TimerQueue that uses a variable-sized array.
struct VariableSizeTimerQueueConfig {
  template <class T>
  struct Container {
    // We need stable references, so use a deque
    typedef std::deque<T> type;
  };

  template <class Timer>
  static Timer* get_timer(typename Container<Timer>::type& timers, typename Container<Timer*>::type& timer_heap,
                          unsigned int timer) {
    if (timer >= timers.size()) {
      timers.resize(timer + 1);
      timer_heap.resize(timer + 1, NULL);
    }
    return &timers[timer];
  }
};

}  // namespace detail

/// A timer queue that has a fixed-size determined at compile-time.
template <class TimerId, std::size_t capacity, class Clock = CocoClock>
class FixedSizeTimerQueue : public detail::TimerQueue<TimerId, Clock, detail::FixedSizeTimerQueueConfig<capacity>> {
 public:
  FixedSizeTimerQueue() = default;
};

/// A timer queue that has a variable-size determined at runtime.
template <class TimerId, class Clock = CocoClock>
class VariableSizeTimerQueue : public detail::TimerQueue<TimerId, Clock, detail::VariableSizeTimerQueueConfig> {
 public:
  VariableSizeTimerQueue() = default;
};

class VariableSizeSingleThreadedTimerExecutor : public SingleThreadedTimerExecutor {
 public:
  void enable_timer(BaseSingleThreadedComponent& component, int timer, CocoClock::duration duration) {
    auto pair =
        timer_map_.insert(std::make_pair(std::make_pair(&component, timer), static_cast<unsigned int>(timers_.size())));
    if (pair.second) {
      timers_.push_back({&component, timer});
    }
    timer_queue_.enable_timer(pair.first->second, duration);
  }

  void cancel_timer(BaseSingleThreadedComponent& component, int timer) {
    timer_queue_.cancel_timer(timer_map_.at({&component, timer}));
  }

 protected:
  bool timer_fired() { return timer_queue_.timer_fired(); }

  void execute_timer() {
    COCO_RUNTIME_ASSERT(timer_fired());
    std::pair<BaseSingleThreadedComponent*, int> timer = timers_[timer_queue_.pop()];
    timer.first->execute_timer(timer.second);
  }

  template <class ConditionVariable, class Lock>
  void wait_until_next_deadline(ConditionVariable& condition_variable, Lock& lock, bool handler_running) {
    timer_queue_.wait_until_next_deadline(condition_variable, lock, handler_running);
  }

 private:
  std::map<std::pair<BaseSingleThreadedComponent*, int>, unsigned int> timer_map_;
  std::vector<std::pair<BaseSingleThreadedComponent*, int>> timers_;
  VariableSizeTimerQueue<unsigned int> timer_queue_;
};

#ifdef COCO_VARIANT_WORKAROUND

namespace detail {

template <typename T, typename... Ts>
struct index_of;

template <typename T, typename... Ts>
struct index_of<T, T, Ts...> {
  static const size_t value = 0;
};

template <typename T, typename S, typename... Ts>
struct index_of<T, S, Ts...> {
  static const size_t value = 1 + index_of<T, Ts...>::value;
};

template <size_t index, typename... Ts>
struct type_of;

template <typename T, typename... Ts>
struct type_of<0, T, Ts...> {
  typedef T type;
};

template <size_t index, typename T, typename... Ts>
struct type_of<index, T, Ts...> {
  typedef typename type_of<index - 1, Ts...>::type type;
};

template <size_t index, typename... Ts>
struct variant_helper;

template <size_t index>
struct variant_helper<index> {
  static inline void destroy(size_t kind, void* storage) {}
  static inline void assign(size_t kind, void* to, const void* from) {}
  static inline void move(const size_t kind, void* to, const void* from) {}
  template <template <typename> class Comparator>
  static inline bool compare(const size_t kind, const void* left, const void* right) {
    abort();
  }
};

template <size_t index, typename T, typename... Ts>
struct variant_helper<index, T, Ts...> {
  static inline void assign(const size_t kind, void* to, const void* from) {
    if (kind == index) {
      new (to) T(*reinterpret_cast<const T*>(from));
    } else {
      variant_helper<index + 1, Ts...>::assign(kind, to, from);
    }
  }

  static inline void move(const size_t kind, void* to, void* from) {
    if (kind == index) {
      new (to) T(std::move(*reinterpret_cast<T*>(from)));
    } else {
      variant_helper<index + 1, Ts...>::move(kind, to, from);
    }
  }

  static inline void destroy(const size_t kind, void* storage) {
    if (kind == index) {
      reinterpret_cast<T*>(storage)->~T();
    } else {
      variant_helper<index + 1, Ts...>::destroy(kind, storage);
    }
  }

  template <template <typename> class Comparator>
  static inline bool compare(const size_t kind, const void* left, const void* right) {
    if (kind == index) {
      return Comparator<T>()(*reinterpret_cast<const T*>(left), *reinterpret_cast<const T*>(right));
    } else {
      return variant_helper<index + 1, Ts...>::template compare<Comparator>(kind, left, right);
    }
  }
};

template <typename... Ts>
struct first;

template <typename T, typename... Ts>
struct first<T, Ts...> {
  typedef T type;
};

}  // namespace detail

template <size_t index>
struct in_place_index_t {
  explicit in_place_index_t() = default;
};

/// Provides support for Coco's tagged enums. Equivalent to C++17's std::variant type.
template <typename... Ts>
class variant {
 public:
  variant() : kind_(0) { new (&storage_) default_type(); }
  variant(const variant& other) { assign(other); }
  variant(variant&& other) COCO_NOEXCEPT { move(std::move(other)); }

  template <class T>
  variant(const T& other) : kind_(detail::index_of<typename std::decay<T>::type, Ts...>::value) {
    new (&storage_) typename std::decay<T>::type(other);
  }

  template <class T, typename Enable = typename std::enable_if<
                         !std::is_same<variant<Ts...>, typename std::decay<T>::type>::value>::type>
  variant(T&& other) : kind_(detail::index_of<typename std::decay<T>::type, Ts...>::value) {
    new (&storage_) typename std::decay<T>::type(std::move(other));
  }

  template <size_t ix, class T>
  variant(in_place_index_t<ix>, const T& other) : kind_(ix) {
    static_assert(0 <= ix && ix < sizeof...(Ts), "index should be between 0 and the number of cases");
    new (&storage_) typename detail::type_of<ix, Ts...>::type(other);
  }

  template <size_t ix, class T>
  variant(in_place_index_t<ix>, T&& other) : kind_(ix) {
    static_assert(0 <= ix && ix < sizeof...(Ts), "index should be between 0 and the number of cases");
    new (&storage_) typename detail::type_of<ix, Ts...>::type(std::move(other));
  }

  ~variant() { destroy(); }

  variant& operator=(const variant& other) {
    destroy();
    assign(other);
    return *this;
  }

  variant& operator=(variant&& other) COCO_NOEXCEPT {
    destroy();
    move(std::move(other));
    return *this;
  }

  std::size_t index() const { return kind_; }

  template <class T>
  T& get() {
    COCO_RUNTIME_ASSERT((detail::index_of<T, Ts...>::value == kind_));
    return *as<T>();
  }

  template <class T>
  const T& get() const {
    COCO_RUNTIME_ASSERT((detail::index_of<T, Ts...>::value == kind_));
    return *as<T>();
  }

  template <size_t ix>
  typename detail::type_of<ix, Ts...>::type& get() {
    COCO_RUNTIME_ASSERT(ix == kind_);
    return *as<typename detail::type_of<ix, Ts...>::type>();
  }

  template <size_t ix>
  const typename detail::type_of<ix, Ts...>::type& get() const {
    COCO_RUNTIME_ASSERT(ix == kind_);
    return *as<typename detail::type_of<ix, Ts...>::type>();
  }

  bool operator==(const variant& right) const {
    return kind_ == right.kind_ &&
           detail::variant_helper<0, Ts...>::template compare<std::equal_to>(kind_, &storage_, &right.storage_);
  }

  bool operator!=(const variant& right) const {
    return kind_ != right.kind_ ||
           detail::variant_helper<0, Ts...>::template compare<std::not_equal_to>(kind_, &storage_, &right.storage_);
  }

 private:
  typedef typename detail::first<Ts...>::type default_type;

  unsigned int kind_;
  typename aligned_union<0, Ts...>::type storage_;

  template <typename T>
  T* as() {
    return reinterpret_cast<T*>(&storage_);
  }
  template <typename T>
  const T* as() const {
    return reinterpret_cast<const T*>(&storage_);
  }

  void assign(const variant& other) {
    kind_ = other.kind_;
    detail::variant_helper<0, Ts...>::assign(kind_, &storage_, &other.storage_);
  }

  void move(variant&& other) {
    kind_ = other.kind_;
    detail::variant_helper<0, Ts...>::move(kind_, &storage_, &other.storage_);
  }

  void destroy() { detail::variant_helper<0, Ts...>::destroy(kind_, &storage_); }
};

template <typename T, typename... Types>
T& get(variant<Types...>& v) {
  return v.template get<T>();
}

template <typename T, typename... Types>
const T& get(const variant<Types...>& v) {
  return v.template get<T>();
}

template <size_t index, typename... Types>
typename detail::type_of<index, Types...>::type& get(variant<Types...>& v) {
  return v.template get<index>();
}

template <size_t index, typename... Types>
const typename detail::type_of<index, Types...>::type& get(const variant<Types...>& v) {
  return v.template get<index>();
}

template <typename... Ts>
bool operator==(const variant<Ts...>& left, const variant<Ts...>& right) {
  return left.operator==(right);
}
template <typename... Ts>
bool operator!=(const variant<Ts...>& left, const variant<Ts...>& right) {
  return left.operator!=(right);
}

/// Provides support for Coco's optional type. Equivalent to C++17's std::optional type.
template <typename T>
class optional {
 public:
  optional() : has_value_(false) {}
  optional(const T& value) : has_value_(true) { new (&storage_) T(value); }
  optional(const optional& other) : has_value_(false) {
    if (other.has_value_) {
      has_value_ = true;
      new (&storage_) T(other.storage());
    }
  }
  optional(T&& value) : has_value_(true) { new (&storage_) T(std::move(value)); }
  optional(optional&& other) COCO_NOEXCEPT : has_value_(false) {
    if (other.has_value_) {
      new (&storage_) T(std::move(other.storage()));
      has_value_ = true;
      other.has_value_ = false;
    }
  }

  ~optional() { reset(); }

  optional<T>& operator=(const optional& other) {
    if (has_value_ && other.has_value_) {
      storage() = other.storage();
    } else if (has_value_) {
      reset();
    } else if (other.has_value_) {
      new (&storage_) T(other.storage());
      has_value_ = true;
    }
    return *this;
  }

  optional<T>& operator=(optional&& other) COCO_NOEXCEPT {
    if (has_value_ && other.has_value_) {
      storage() = std::move(other.storage());
    } else if (has_value_) {
      reset();
    } else if (other.has_value_) {
      new (&storage_) T(std::move(other.storage()));
      has_value_ = true;
    }
    return *this;
  }

  optional<T>& operator=(const T& other) {
    if (has_value_) {
      storage() = other;
    } else {
      has_value_ = true;
      new (&storage_) T(other);
    }
    return *this;
  }

  optional<T>& operator=(T&& other) {
    if (has_value_) {
      storage() = std::move(other);
    } else {
      has_value_ = true;
      new (&storage_) T(std::move(other));
    }
    return *this;
  }

  bool has_value() const { return has_value_; }

  T& value() {
    COCO_RUNTIME_ASSERT(has_value_);
    return storage();
  }

  const T& value() const {
    COCO_RUNTIME_ASSERT(has_value_);
    return storage();
  }

  T value_or(T&& default_value) const { return has_value_ ? storage() : std::move(default_value); }

  void reset() {
    if (has_value_) {
      storage().~T();
      has_value_ = false;
    }
  }

 private:
  typename aligned_union<0, T>::type storage_;
  bool has_value_;

  T& storage() { return *reinterpret_cast<T*>(&storage_); }
  const T& storage() const { return *reinterpret_cast<const T*>(&storage_); }
};

template <typename T>
bool operator==(const optional<T>& left, const optional<T>& right) {
  return left.has_value() == right.has_value() && (!left.has_value() || left.value() == right.value());
}

template <typename T>
bool operator!=(const optional<T>& left, const optional<T>& right) {
  return left.has_value() != right.has_value() || (left.has_value() && left.value() != right.value());
}

#else

template <typename... Ts>
using variant = std::variant<Ts...>;
template <size_t index>
using in_place_index_t = std::in_place_index_t<index>;

template <class T, class... Types>
constexpr T& get(variant<Types...>& v) {
  return ::std::get<T>(v);
}
template <class T, class... Types>
constexpr T&& get(variant<Types...>&& v) {
  return ::std::get<T>(std::move(v));
}
template <class T, class... Types>
constexpr const T& get(const variant<Types...>& v) {
  return ::std::get<T>(v);
}
template <class T, class... Types>
constexpr const T&& get(const variant<Types...>&& v) {
  return ::std::get<T>(std::move(v));
}

template <std::size_t index, class... Types>
constexpr std::variant_alternative_t<index, variant<Types...>>& get(variant<Types...>& v) {
  return ::std::get<index>(v);
}

template <std::size_t index, class... Types>
constexpr std::variant_alternative_t<index, variant<Types...>>&& get(variant<Types...>&& v) {
  return ::std::get<index>(std::move(v));
}

template <std::size_t index, class... Types>
constexpr const std::variant_alternative_t<index, variant<Types...>>& get(const variant<Types...>& v) {
  return ::std::get<index>(v);
}

template <std::size_t index, class... Types>
constexpr const std::variant_alternative_t<index, variant<Types...>>&& get(const variant<Types...>&& v) {
  return ::std::get<index>(std::move(v));
}

template <typename T>
using optional = std::optional<T>;

#endif

/// A wrapper of type T that causes value-initialisation to initialise the value to the given default value.
template <typename T, T DefaultValue>
struct initialized {
  initialized() : value(DefaultValue) {}
  initialized(const T& value) : value(value) {}
  initialized(T&& value) : value(std::move(value)) {}
  initialized(const initialized& other) : value(other.value) {}
  initialized(initialized&& other) COCO_NOEXCEPT : value(std::move(other.value)) {}

  initialized& operator=(const T& other) {
    value = other;
    return *this;
  }
  initialized& operator=(T&& other) {
    value = std::move(other);
    return *this;
  }
  initialized& operator=(const initialized& other) {
    value = other.value;
    return *this;
  }
  initialized& operator=(initialized&& other) COCO_NOEXCEPT {
    value = std::move(other.value);
    return *this;
  }

  operator T&() { return value; }
  operator const T&() const { return value; }

  T value;
};

/// A function that when called simply returns the specified value.
template <typename T, T value>
T return_constant() {
  return value;
}

/// Invokes a member function on each member of the given container.
///
/// It is logically equivalent to:
///   std::for_each(items.begin(), items.end(), std::bind(fn, __1, args...))
/// but has the advantage of reducing the size of the generated assembly code.
template <class Container, typename MemberFunction, typename... Args>
void invoke_all(const Container& items, MemberFunction fn, Args&&... args) {
  for (auto it = items.begin(), it_end = items.end(); it != it_end; ++it) {
    ((*it)->*fn)(std::forward<Args>(args)...);
  }
}

/// A helper class for coco::expected which can be used to return an error from a function that returns expected<T, E>.
///
/// @seealso expected
template <typename E>
class unexpected {
 public:
  unexpected() = delete;
  unexpected(const E& error) : error_(error) {}
  unexpected(E&& error) : error_(std::move(error)) {}
  unexpected(const unexpected&) = default;
  unexpected& operator=(const unexpected&) = default;
  unexpected(unexpected&& other) COCO_NOEXCEPT : error_(std::move(other.error_)) {}
  unexpected& operator=(unexpected&& other) COCO_NOEXCEPT {
    error_ = std::move(other.error_);
    return *this;
  }
  ~unexpected() = default;

  E& value() { return error_; }
  const E& value() const { return error_; }

 private:
  E error_;
};

template <class E>
unexpected<typename std::decay<E>::type> make_unexpected(E&& e) {
  return unexpected<typename std::decay<E>::type>(std::forward<E>(e));
}

/// A simple P0323-like implementation of std::expected, suitable for coco's Result type.
///
/// Suppose you have a function such as:
/// \code
///   coco::expected<std::vector<bool>, int> mayFail();
/// \endcode
/// Within the body of mayFail, you can use:
/// \code
///   return coco::make_unexpected(5);
/// \endcode
/// to return a specific error code, or:
/// \code
///   return std::vector<bool>()
/// \endcode
/// to return the expected value.
///
/// When calling mayFail(), the accessors has_value(), value() and error() can be used to either access the value or
/// the container error, as required.
///
/// @tparam T the type of a success case.
/// @tparam E the type of an error.
template <typename T, typename E>
class expected {
 public:
  expected() : storage_() {}
  expected(const expected&) = default;
  expected& operator=(const expected&) = default;
  expected(expected&& other) COCO_NOEXCEPT : storage_(std::move(other.storage_)) {}
  expected& operator=(expected&& other) COCO_NOEXCEPT {
    storage_ = std::move(other.storage_);
    return *this;
  }
  ~expected() = default;

  explicit expected(const T& value) : storage_(in_place_index_t<0>{}, value) {}
  expected(T&& value) : storage_(in_place_index_t<0>{}, std::move(value)) {}
  expected(const unexpected<E>& error) : storage_(in_place_index_t<1>{}, error.value()) {}
  expected(unexpected<E>&& error) : storage_(in_place_index_t<1>{}, std::move(error.value())) {}

  /// Returns true if this expected contains an error rather than a value.
  bool has_error() const { return storage_.index() == 1; }

  /// Returns true if this expected contains a value rather than an error.
  bool has_value() const { return storage_.index() == 0; }

  /// If has_value is false, then returns the error.
  E& error() { return coco::get<1>(storage_); }
  const E& error() const { return coco::get<1>(storage_); }

  /// If has_value is true, then returns the error.
  T& value() { return coco::get<0>(storage_); }
  const T& value() const { return coco::get<0>(storage_); }

  T value_or(T&& value) { return has_value() ? coco::get<0>(storage_) : value; }

  bool operator==(const expected<T, E>& right) const { return storage_ == right.storage_; }
  bool operator!=(const expected<T, E>& right) const { return storage_ != right.storage_; }

 private:
  variant<T, E> storage_;
};

template <typename T, typename E>
bool operator==(const expected<T, E>& left, const expected<T, E>& right) {
  return left.operator==(right);
}

template <typename T, typename E>
bool operator!=(const expected<T, E>& left, const expected<T, E>& right) {
  return left.operator!=(right);
}

/// An empty type to properly replicate Coco's Nil type.
struct Void {};

template <class ProvidedPort>
class AutoShutdown {
 public:
  AutoShutdown(ProvidedPort& port) : port(port) { port.register_client(); }
  ~AutoShutdown() { port.unregister_client(); }

 private:
  ProvidedPort& port;
};

}  // namespace v1
}  // namespace coco

#undef COCO_VARIANT_WORKAROUND
#undef COCO_ALWAYS_ASSERT
#undef COCO_GCC44_WORKAROUNDS
#undef COCO_GCC48_WORKAROUNDS

#endif
