#ifndef COCO_STREAM_LOGGER_H_
#define COCO_STREAM_LOGGER_H_

// MIT License
//
// Copyright (c) 2019 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include <iostream>
#include <mutex>

#include "coco/runtime.h"

namespace coco {

/// A simple logger that outputs to a std::ostream.
///
/// This is not intended for production usage, but as an example of how a logger could be implemented.
class StreamMachineLogger : public StateMachineLogger {
 public:
  /// Returns a pointer to a shared logger that logs to stdout.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cout);
  /// @endcode
  static StreamMachineLogger* cout() {
    static StreamMachineLogger logger(std::cout);
    return &logger;
  }

  /// Returns a pointer to a shared logger that logs to stderr.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cerr);
  /// @endcode
  static StreamMachineLogger* cerr() {
    static StreamMachineLogger logger(std::cerr);
    return &logger;
  }

  StreamMachineLogger(std::ostream& out) : out(out) {}

  void machine_starting(const BaseComponent& component) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << ": starting machine" << std::endl;
  }

  void machine_started(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << current_state << "): started machine" << std::endl;
  }

  void machine_terminated(const BaseComponent& component) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << ": terminated machine" << std::endl;
  }

  void state_entry_started(const BaseComponent& component, const char* state) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << state << "): entry started" << std::endl;
  }

  void state_exit_finished(const BaseComponent& component, const char* state) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << state << "): exit finished" << std::endl;
  }

  void event_started(const BaseComponent& component, const char* current_state, const char* port,
                     const char* event) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << current_state << "): started " << port << "." << event << std::endl;
  }
  void event_finished(const BaseComponent& component, const char* current_state, const char* port,
                      const char* event) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << current_state << "): finished " << port << "." << event << std::endl;
  }

  void message(const BaseComponent& component, const char* current_state, const char* message) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex);
    out << component.name() << " (" << current_state << "): " << message << std::endl;
  }

  void abort(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    message(component, current_state, "abort");
  }

 private:
  std::mutex mutex;
  std::ostream& out;
};

}  // namespace coco

#endif
